
import sys
import re
import pickle
import numpy as np

CLASS_1 = 1  # next is space
CLASS_0 = 0  # next is not space

S1 = re.compile(u'''[\s]+''')


def snorm(string):
    return S1.sub(' ', string.replace('\t', ' ')).strip()


def open_file(filename, mode):
    try:
        fid = open(filename, mode, encoding="utf-8")
    except:
        sys.stderr.write("open_file(), file open error : %s\n" % (filename))
        exit(1)
    else:
        return fid


def close_file(fid):
    fid.close()


def build_dictionary(train_path, padd):
    char_rdic = []
    visit = {}
    fid = open_file(train_path, 'r')
    for line in fid:
        line = line.strip()
        if line == "": continue
        for c in line:
            if c not in visit:
                char_rdic.append(c)
                visit[c] = 1
    if padd not in visit: char_rdic.append(padd)
    char_dic = {w: i for i, w in enumerate(char_rdic)}  # char to id
    close_file(fid)
    return char_dic


def one_hot(i, size):
    return [1 if j == i else 0 for j in range(size)]


def get_xy_data(sentence, pos, n_steps, padd):
    slen = len(sentence)
    x_data = []
    y_data = []
    next_pos = -1
    count = 0
    i = pos
    while i < slen:
        c = sentence[i]
        x_data.append(c)
        next_c = None
        if i + 1 < slen: next_c = sentence[i + 1]
        if next_c == ' ':
            y_data.append(CLASS_1)
        else:
            y_data.append(CLASS_0)
        count += 1
        i += 1
        if count == n_steps: break
    if count == n_steps:
        if i == slen:
            # reached end
            next_pos = -1
        if i < slen:
            # move prev space + 1
            j = i - 1
            space_count = 0
            while j > 0:
                c = sentence[j]
                if c == ' ':
                    space_count += 1
                    if space_count == 1: break
                if i - j >= 12:  # no prev space atmost 12
                    break
                j -= 1
            if j <= i - 1:
                next_pos = j + 1
    else:
        # padding
        diff = n_steps - count
        x_data += [padd] * diff
        y_data += [CLASS_0] * diff
        next_pos = -1
    return x_data, y_data, next_pos, count


def next_batch(sentence, pos, char_dic, vocab_size, n_steps, padd):
    '''
    y_data =  1 or 0     => n_steps unfolding => [0,0,1,0,...]
    ^
    |
    x_data = [1,0,...,0] => n_steps unfolding => [[1,0,..0],..,[0,0,1,..0]]

    batch_xs.shape => (batch_size=1, n_steps, n_input)
    batch_ys.shape => (batch_size=1, n_steps)
    '''
    batch_xs = []
    batch_ys = []
    x_data, y_data, next_pos, count = get_xy_data(sentence, pos, n_steps, padd)
    x_data = [char_dic[c] if c in char_dic else padd for c in x_data]
    x_data = [one_hot(i, vocab_size) for i in x_data]
    batch_xs.append(x_data)
    batch_ys.append(y_data)
    batch_xs = np.array(batch_xs, dtype='f')
    batch_ys = np.array(batch_ys, dtype='int32')
    return batch_xs, batch_ys, next_pos, count


def test_next_batch(train_path, char_dic, vocab_size, n_steps, padd):
    fid = open_file(train_path, 'r')
    for line in fid:
        line = line.strip()
        if line == "": continue
        line = line.decode('utf-8')
        sentence = snorm(line)
        pos = 0
        while pos != -1:
            batch_xs, batch_ys, next_pos, count = next_batch(sentence, pos, char_dic, vocab_size, n_steps, padd)
            print
            'window : ' + sentence[pos:pos + n_steps].encode('utf-8')
            print
            'count : ' + str(count)
            print
            'next_pos : ' + str(next_pos)
            print
            batch_ys
            pos = next_pos
    close_file(fid)


def get_validation_data(validation_path, char_dic, vocab_size, n_steps, padd):
    validation_data = []
    fid = open_file(validation_path, 'r')
    for line in fid:
        line = line.strip()
        if line == "": continue
        sentence = snorm(line)
        pos = 0
        while pos != -1:
            batch_xs, batch_ys, next_pos, count = next_batch(sentence, pos, char_dic, vocab_size, n_steps, padd)
            validation_data.append((batch_xs, batch_ys, count))
            pos = next_pos
    close_file(fid)
    return validation_data


def to_sentence(tag_vector, sentence):
    out = []
    j = 0
    tag_vector_size = len(tag_vector)
    sentence_size = len(sentence)
    while j < tag_vector_size and j < sentence_size:
        tag = tag_vector[j]
        if tag == CLASS_1:
            out.append(sentence[j])
            if sentence[j] != ' ': out.append(' ')
        else:
            out.append(sentence[j])
        j += 1
    n_sentence = ''.join(out)
    return snorm(n_sentence)


# -------------------------------------------------------------------------

def build_dictionary_emb(embedding_dir):
    embedding_path = embedding_dir + '/' + 'embedding.pickle'
    with open(embedding_path, 'rb') as handle:
        embeddings = pickle.load(handle)
    id2ch = {}
    vocab_path = embedding_dir + '/' + 'vocab.txt'
    idx = 0
    with open(vocab_path, "r") as handle:
        for line in handle:
            ch, count = line.split(' ')
            id2ch[idx] = ch.decode('utf-8')
            idx += 1
    assert (idx == len(embeddings))
    ch2id = {}
    id2emb = {}
    for i, ch in id2ch.iteritems():
        ch2id[ch] = i
        try:
            id2emb[i] = embeddings[i]
        except Exception as e:
            sys.stderr.write('vocab.txt, embedding.pickle not aligned\n')
            sys.exit(1)
    embedding_dim = len(embeddings[0])
    return ch2id, id2ch, id2emb, embedding_dim


def next_batch_emb(sentence, pos, char_dic, id2emb, n_steps, padd):
    '''
    y_data =  1 or 0     => n_steps unfolding => [0,0,1,0,...]
    ^
    |
    x_data = [1,0,...,0] => n_steps unfolding => [[1.0,1.4,..3.2],..,[0.2,0.6,3.7,..,0.0]]

    batch_xs.shape => (batch_size=1, n_steps, n_input)
    batch_ys.shape => (batch_size=1, n_steps)
    '''
    batch_xs = []
    batch_ys = []
    x_data, y_data, next_pos, count = get_xy_data(sentence, pos, n_steps, padd)
    tmp_x_data = []
    for c in x_data:
        id = 0  # for 'UNK'
        if c in char_dic: id = char_dic[c]
        '''
        print 'c, id = ', c.encode('utf-8'), id
        print id2emb[id]
        '''
        tmp_x_data.append(id2emb[id])
    x_data = tmp_x_data
    batch_xs.append(x_data)
    batch_ys.append(y_data)
    batch_xs = np.array(batch_xs, dtype='f')
    batch_ys = np.array(batch_ys, dtype='int32')
    return batch_xs, batch_ys, next_pos, count


def test_next_batch_emb(train_path, char_dic, id2emb, n_steps, padd):
    fid = open_file(train_path, 'r')
    for line in fid:
        line = line.strip()
        if line == "": continue
        line = line.decode('utf-8')
        sentence = snorm(line)
        pos = 0
        while pos != -1:
            batch_xs, batch_ys, next_pos, count = next_batch_emb(sentence, pos, char_dic, id2emb, n_steps, padd)
            print
            'window : ' + sentence[pos:pos + n_steps].encode('utf-8')
            print
            'count : ' + str(count)
            print
            'next_pos : ' + str(next_pos)
            print
            batch_xs
            print
            batch_ys
            pos = next_pos
    close_file(fid)


def get_validation_data_emb(validation_path, char_dic, id2emb, n_steps, padd):
    validation_data = []
    fid = open_file(validation_path, 'r')
    for line in fid:
        line = line.strip()
        if line == "": continue
        line = line.decode('utf-8')
        sentence = snorm(line)
        pos = 0
        while pos != -1:
            batch_xs, batch_ys, next_pos, count = next_batch_emb(sentence, pos, char_dic, id2emb, n_steps, padd)
            validation_data.append((batch_xs, batch_ys, count))
            pos = next_pos
    close_file(fid)
    return validation_data



class Candidate(str):
    def __init__(self, value):
        super().__init__()
        self.start = 0
        self.stop = 0

    def set_position(self, start, stop):
        self.start = start
        self.stop = stop



from sklearn.model_selection import train_test_split

from treform.keyword.ml_keyword_extractor import extract_candidate_keywords, extract_candidate_keywords_for_training
from nltk.classify.scikitlearn import SklearnClassifier
from nltk.metrics.scores import accuracy, precision, recall, f_measure
from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVC,LinearSVC
from random import shuffle

import pandas as pd
from joblib import dump, load
import collections
import os
import re
import nltk
import pprint
import time
import numpy as np
from treform.tokenizer import Komoran
#Print features with nice indentation
pp = pprint.PrettyPrinter(indent=4)

#Extract features of a given word
def get_features(document, candidate, candidate_list, isKeyword, language='en'):
    features = {}
    features['length'] = len(candidate.split(' '))
    if language == 'en':
        features['part_of_speech'] = ' '.join([pos for word,pos in nltk.pos_tag(nltk.word_tokenize(candidate))])
    elif language == 'ko':
        komoran = Komoran()
        features['part_of_speech'] = ' '.join([pos for word, pos in komoran(candidate)])
    position_list = [ m.start()/float(len(document)) for m in re.finditer(re.escape(candidate),document,flags=re.IGNORECASE)]

    if len(position_list):
        for i in range(0,len(position_list)):
            features[str(i) + 'th occrrence'] = position_list[i]

    N = len(document)
    #Line approximation

    if len(position_list):
        y = 0
        for pos in position_list:
            #x = document.index(candidate)
            x = pos*N
            if x >= float(N)/2:
                y += 2/float(N) * x - 1
            else :
                y += -2/float(N) * x + 1

        y /= len(position_list)
    else:
        x = 0
        y = 0

    features['line_position'] = y

    #Feature: Try the same approximation with parabola
    if len(position_list):
        y = 0
        for pos in position_list:
                x = pos*N
                y += ((x - float(N)/2)**2 )/ ((float(N)**2) / 4)
        y /= len(position_list)
    else:
        x = 0
        y = 0

    features['parabolic_position'] = y

    #Feature: Standard deviation
    if len(position_list):
        avg_position = sum([pos*float(len(document)) for pos in position_list])/float(len(position_list))
        ss = 0.0
        for p in position_list:
            ss += (p*float(len(document))  - avg_position)**2
        ss /= len(position_list)
        ss = ss**0.5
        features['Standard deviation'] = ss/float(len(document))

    #Feature: Text Frequency
    features['frequency'] = len(position_list)/ float(len(set(candidate_list))) #check

    pp.pprint(features)
    print('for the top : ' + str(isKeyword))
    return features

# Determine Precision, Recall & F-Measure
def accuracy_measure(classifier,cross_valid_set):
    refsets = collections.defaultdict(set)
    testsets = collections.defaultdict(set)

    for i, (feats, label) in enumerate(cross_valid_set):
        refsets[label].add(i)
        observed = classifier.classify(feats)
        testsets[observed].add(i)

    print('pos Precision:', precision(refsets[1], testsets[1]))
    print('pos Recall:', recall(refsets[1], testsets[1]))
    print('pos F-measure:', f_measure(refsets[1], testsets[1]))
    print('neg Precision:', precision(refsets[0], testsets[0]))
    print('neg Recall:', recall(refsets[0], testsets[0]))
    print('neg F-measure:', f_measure(refsets[0], testsets[0]))

def train(language='ko', doc_path=None, key_path=None,model_path=None):
    from sklearn import metrics
    # Initialize data and target lists
    target = []
    data = []
    feature_list = []

    start_time = time.time()

    #language = 'ko'  # either en or ko

    if language == 'en':
        if doc_path == None:
            doc_path = '../../sample_data/keyphrase_training_data/english/documents/'
        if key_path == None:
            key_path = '../../sample_data/keyphrase_training_data/english/teams/'

        for doc in os.listdir(doc_path):
            doc_name = doc.split('.')[0]

            if doc.endswith('txt'):
                document = open(doc_path + doc, "r", encoding='utf-8').read()

                # Get List of candidates
                candidates = extract_candidate_keywords_for_training(document, language='en')

                # #Initialize data and target lists
                print('Now going to ', doc)
                keyword_list = []
                for team in os.listdir(key_path):
                    if team.startswith('team'):
                        keywords = open(key_path + team + '/' + doc_name + '.key', "r", encoding='utf-8').read()
                        # Take list of Keywords
                        keyword_list.extend(
                            [line.split(':')[1].lower().strip() for line in keywords.splitlines() if ':' in line])

                feature_list.extend([(get_features(document, key, candidates, 1, language='ko'), 1) for key in set(keyword_list)])
                feature_list.extend([(get_features(document, key, candidates, 0, language='ko'), 0) for key in candidates if key not in set(keyword_list)])

                print(len(feature_list))

    elif language == 'ko':
        if doc_path == None:
            doc_path = '../../sample_data/keyphrase_training_data/korean/KeaRaw'
        if key_path == None:
            key_path = '../../sample_data/keyphrase_training_data/korean/KeaTrain'

        k_arr = os.listdir(key_path)
        k_tup = {}
        for a_file in k_arr:
            if a_file.endswith('.key'):
                abs_file = os.path.join(key_path, a_file)
                content = open(abs_file, 'r', encoding='utf-8').readlines()
                # print(content)
                prefix = a_file.strip('.key')
                k_tup[prefix] = content

        arr = os.listdir(doc_path)
        tup = {}
        for i, a_file in enumerate(arr):
            prefix = a_file.strip('.txt')
            # print(prefix)
            #print(doc_path)
            #print(a_file)
            abs_file = os.path.join(doc_path, a_file)
            content = open(abs_file, 'r', encoding='utf-8').read()
            # print(content)
            tup[prefix] = content
            k_content = k_tup[prefix]
            keyword_list = []
            for keyword in k_content:
                keyword_list.append(keyword)

            candidates = extract_candidate_keywords_for_training(content, language='ko')
            if len(candidates) > 0:
                feature_list.extend([(get_features(content, key, candidates, 1, language='ko'), 1) for key in set(keyword_list)])
                feature_list.extend([(get_features(content, key, candidates, 0, language='ko'), 0) for key in candidates if
                                     key not in set(keyword_list)])

    print(len(feature_list))
    end_time = time.time()

    shuffle(feature_list)

    # split
    good_size = len([(x, y) for x, y in feature_list if y == 1])
    bad_size = len([(x, y) for x, y in feature_list if y == 0])
    size = len(feature_list)
    good_train_size = int(good_size * 0.7)
    good_test_size = good_size - good_train_size

    bad_train_size = int(bad_size * 0.7)
    bad_test_size = bad_size - bad_train_size

    good_list_train = [(x, y) for x, y in feature_list if y == 1][:good_train_size]
    good_list_test = [(x, y) for x, y in feature_list if y == 1][good_train_size:]
    bad_list_train = [(x, y) for x, y in feature_list if y == 0][:bad_train_size]
    bad_list_test = [(x, y) for x, y in feature_list if y == 0][bad_train_size:]

    print(len(good_list_train))
    print(len(good_list_test))
    print(len(bad_list_train))
    print(len(bad_list_test))

    shuffle(good_list_train)
    shuffle(bad_list_train)
    shuffle(good_list_test)
    shuffle(bad_list_test)

    print('number of keywords', len([x for x, y in feature_list if y == 1]))
    print('number of non keywords', len([x for x, y in feature_list if y == 0]))

    training_set = []
    training_set.extend(good_list_train)
    training_set.extend(bad_list_train)
    print('all indicies of keywords training', len([x for x, y in training_set if y == 1]))

    testing_set = []
    testing_set.extend(good_list_test)
    testing_set.extend(bad_list_test)
    print('all indicies of keywords testing', len([x for x, y in testing_set if y == 1]))

    shuffle(training_set)
    shuffle(testing_set)

    # Logistic Regression
    LogisticRegression_classifier = SklearnClassifier(LogisticRegression()).train(training_set)
    print("LogisticRegression_classifier algorithm accuracy : ",
          nltk.classify.accuracy(LogisticRegression_classifier, testing_set) * 100)
    accuracy_measure(LogisticRegression_classifier, testing_set)

    # SVM
    LinearSVC_classifier = SklearnClassifier(LinearSVC()).train(training_set)
    print("LinearSVC_classifier algorithm accuracy : ", nltk.classify.accuracy(LinearSVC_classifier, testing_set) * 100)
    accuracy_measure(LinearSVC_classifier, testing_set)

    if model_path == None:
        model_path = '../../models/svm_keyphrase.model'
    dump(LinearSVC_classifier, model_path)

    print('Time of execution is %s seconds!' % (end_time - start_time))

if __name__ == '__main__':
    train()


import nltk
import sklearn
from sklearn.feature_extraction.text import TfidfVectorizer
import numpy as np
import pandas as pd
import os
import re
from joblib import dump, load
from sklearn.naive_bayes import MultinomialNB
from sklearn.ensemble import RandomForestClassifier
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import accuracy_score
from sklearn.model_selection import train_test_split
from sklearn.svm import LinearSVC
import treform as ptm

class MachineLearningSentimentAnalyzer:

    def __init__(self):
        name = 'MLSentimentAnalyzer'
        self.reviews_trin_clean = []
        self.reviews_test_clean = []
        self.target=[]

    def split(self):
        reviews_train = []
        for line in open('../data/movie_data/full_train.txt', 'r'):
            reviews_train.append(line.strip())

        reviews_test = []
        for line in open('../data/movie_data/full_test.txt', 'r'):
            reviews_test.append(line.strip())

        self.target = [1 if i < 12500 else 0 for i in range(25000)]

        self.reviews_train_clean = self.preprocess(reviews_train)
        self.reviews_test_clean = self.preprocess(reviews_test)

    def preprocess(self, corpus, language='ko'):
        pipeline = None

        if language == 'ko':
            mecab_path = 'C:\\mecab\\mecab-ko-dic'
            pipeline = ptm.Pipeline(ptm.splitter.NLTK(),
                                    ptm.tokenizer.MeCab(mecab_path),
                                    ptm.helper.POSFilter('NN*'),
                                    ptm.helper.SelectWordOnly(),
                                    ptm.ngram.NGramTokenizer(1, 2),
                                    ptm.helper.StopwordFilter(file='../stopwords/stopwordsKor.txt')
                                    )
        elif language == 'en':
            pipeline = ptm.Pipeline(ptm.splitter.NLTK(),
                                    ptm.tokenizer.WordPos(),
                                    ptm.helper.POSFilter('NN*'),
                                    ptm.helper.SelectWordOnly(),
                                    ptm.ngram.NGramTokenizer(1, 2),
                                    ptm.helper.StopwordFilter(file='../stopwords/stopwordsEng.txt')
                                    )
        result = pipeline.processCorpus(corpus)
        print('==  ==')

        documents = []
        for doc in result:
            document = ''
            for sent in doc:
                document += " ".join(sent)
            documents.append(document)

        return documents

    def at_least_one(self, without_sw, train):
        threshold = len(train) * 0.01
        reqlist = {k: v for k, v in without_sw.items() if v >= threshold}
        return reqlist

    def at_least_twice(self, words_pos, words_neg, pos_total_dict, neg_total_dict):
        # final_dict_pos={k:v for k,v in words_pos.i
        final_dict = {}
        # print len(words_pos)
        # print len(words_neg)

        for word in words_pos.keys():
            if not word in neg_total_dict:
                final_dict[word] = 1
            elif words_pos[word] >= 2 * neg_total_dict[word]:
                final_dict[word] = 1

        # print len(final_dict)
        for word in words_neg.keys():
            if not word in pos_total_dict:
                final_dict[word] = 1
            elif words_neg[word] >= 2 * pos_total_dict[word]:
                final_dict[word] = 1
        return final_dict

    def remove_stopwords(self, total_dict, stopwords):
        filtereddict = {k: v for k, v in total_dict.items() if k not in stopwords}
        return filtereddict

    def make_customized_feature_vectors(self, train_pos, train_neg, test_pos, test_neg):
        """
        Returns the feature vectors for all text in the train and test datasets.
        """
        # Determine a list of words that will be used as features.
        # This list should have the following properties:
        #   (1) Contains no stop words
        #   (2) Is in at least 1% of the positive texts or 1% of the negative texts
        #   (3) Is in at least twice as many postive texts as negative texts, or vice-versa.
        pos_total_dict = {}
        for line in train_pos:
            pos_word_dict = {}
            for word in line:
                pos_word_dict[word] = 1
            for word in pos_word_dict.keys():
                if word in pos_total_dict:
                    pos_total_dict[word] = pos_total_dict[word] + 1
                else:
                    pos_total_dict[word] = 1

        neg_total_dict = {}
        for line in train_neg:
            neg_word_dict = {}
            for word in line:
                neg_word_dict[word] = 1
            for word in neg_word_dict.keys():
                if word in neg_total_dict:
                    neg_total_dict[word] = neg_total_dict[word] + 1
                else:
                    neg_total_dict[word] = 1
        print(len(pos_total_dict))
        print(len(neg_total_dict))

        at_least_one_words_pos = self.at_least_one(pos_total_dict, train_pos)
        at_least_one_words_neg = self.at_least_one(neg_total_dict, train_neg)
        # print len(at_least_one_words_pos)
        # print len(at_least_one_words_neg)
        final_dict = self.at_least_twice(at_least_one_words_pos, at_least_one_words_neg, pos_total_dict, neg_total_dict)

        # Using the above words as features, construct binary vectors for each text in the training and test set.
        # These should be python lists containing 0 and 1 integers.
        feature_list = final_dict.keys()

        train_pos_vec = []
        train_neg_vec = []
        test_pos_vec = []
        test_neg_vec = []

        for line in train_pos:
            line_dict = {}
            for word in line:
                line_dict[word] = 1
            vector_list = []
            for word in feature_list:
                if word in line_dict:
                    vector_list.append(1)
                else:
                    vector_list.append(0)
            train_pos_vec.append(vector_list)

        for line in train_neg:
            line_dict = {}
            for word in line:
                line_dict[word] = 1
            vector_list = []
            for word in feature_list:
                if word in line_dict:
                    vector_list.append(1)
                else:
                    vector_list.append(0)
            train_neg_vec.append(vector_list)

        for line in test_pos:
            line_dict = {}
            for word in line:
                line_dict[word] = 1
            vector_list = []
            for word in feature_list:
                if word in line_dict:
                    vector_list.append(1)
                else:
                    vector_list.append(0)
            test_pos_vec.append(vector_list)

        for line in test_neg:
            line_dict = {}
            for word in line:
                line_dict[word] = 1
            vector_list = []
            for word in feature_list:
                if word in line_dict:
                    vector_list.append(1)
                else:
                    vector_list.append(0)
            test_neg_vec.append(vector_list)

        # Return the four feature vectors
        return train_pos_vec, train_neg_vec, test_pos_vec, test_neg_vec

    def make_baseline_feature_vectors(self, combined_vectors, labels):
        baseline_vectorizer = CountVectorizer()
        features = baseline_vectorizer.fit_transform(combined_vectors)

        self.saveVectorizer(baseline_vectorizer)

        #print('feature size ' + str(len(features)) + " : " + str(len(labels)))

        X_train, X_test, y_train, y_test = train_test_split(features, labels,
                                                            test_size=0.33, random_state=0)

        return X_train, X_test, y_train, y_test

    def make_tfidf_feature_vectors(self, combined_vectors, labels):
        tfidf_vectorizer = TfidfVectorizer()
        features = tfidf_vectorizer.fit_transform(combined_vectors)
        self.saveVectorizer(tfidf_vectorizer)

        #print('feature size ' + " : " + str(len(labels)))

        X_train, X_test, y_train, y_test = train_test_split(features, labels,
                                                            test_size=0.33, random_state=0)

        return X_train, X_test, y_train, y_test

    def make_feature_vectors(self, features, labels):
        '''
        :param features:
        :param labels:
        :return X_train, X_test, y_train, y_test:
        '''
        print('feature size ' + str(len(features)) + " : " + str(len(labels)))
        X_train, X_test, y_train, y_test = train_test_split(features, labels,
                                                            test_size=0.33, random_state=0)

        return X_train, X_test, y_train, y_test

    def build_baseline_model(self, train_X, train_Y, algorithm):
        """
        Returns a ML-based Model that are fit to the training data.
        """
        print(train_X[:10])

        model = None
        if algorithm == 'lr':
            logr_model = LogisticRegression(random_state=0)
            model = logr_model.fit(train_X, train_Y)
        elif algorithm == 'nb':
            gnb = MultinomialNB()
            model = gnb.fit(train_X, train_Y)
        elif algorithm == 'rf':
            rf_classifier = RandomForestClassifier(n_estimators=200, max_depth=3, random_state=0)
            model = rf_classifier.fit(train_X, train_Y)
        elif algorithm == 'svm':
            svm_classifier = LinearSVC()
            model = svm_classifier.fit(train_X, train_Y)

        return model

    def build_models_with_customizedvectors(self, train_pos_X, train_neg_X, algorithm):
        train_Y = ["pos"] * train_pos_X.shape[1] + ["neg"] * train_neg_X.shape[1]
        train_X = train_pos_X + train_neg_X
        print(train_X[:10])

        model = None
        if algorithm == 'lr':
            logr_model = LogisticRegression(random_state=0)
            model = logr_model.fit(train_X, train_Y)
        elif algorithm == 'nb':
            gnb = MultinomialNB()
            model = gnb.fit(train_X, train_Y)
        elif algorithm == 'rf':
            rf_classifier = RandomForestClassifier(n_estimators=200, max_depth=3, random_state=0)
            model = rf_classifier.fit(train_X, train_Y)
        elif algorithm == 'svm':
            svm_classifier = LinearSVC()
            model = svm_classifier.fit(train_X, train_Y)

        return model

    def baselineLR(self):
        baseline_vectorizer = CountVectorizer(binary=True)
        baseline_vectorizer.fit(self.reviews_train_clean)
        X_baseline = baseline_vectorizer.transform(self.reviews_train_clean)
        X_test_baseline = baseline_vectorizer.transform(self.reviews_test_clean)

        X_train, X_val, y_train, y_val = train_test_split(
            X_baseline, self.target, train_size=0.75)

        for c in [0.01, 0.05, 0.25, 0.5, 1]:
            lr = LogisticRegression(C=c)
            lr.fit(X_train, y_train)
            print("Accuracy for C=%s: %s"
                  % (c, accuracy_score(y_val, lr.predict(X_val))))

        final_model = LogisticRegression(C=0.05)
        final_model.fit(X_baseline, self.target)
        print("Final Accuracy: %s"
              % accuracy_score(self.target, final_model.predict(X_test_baseline)))
        # Final Accuracy: 0.88128

    def tfidfLR(self):
        tfidf_vectorizer = TfidfVectorizer()
        tfidf_vectorizer.fit(self.reviews_train_clean)
        X = tfidf_vectorizer.transform(self.reviews_train_clean)
        X_test = tfidf_vectorizer.transform(self.reviews_test_clean)

        X_train, X_val, y_train, y_val = train_test_split(
            X, self.target, train_size=0.75)

        for c in [0.01, 0.05, 0.25, 0.5, 1]:
            lr = LogisticRegression(C=c)
            lr.fit(X_train, y_train)
            print("Accuracy for C=%s: %s"
                  % (c, accuracy_score(y_val, lr.predict(X_val))))

        final_tfidf = LogisticRegression(C=1)
        final_tfidf.fit(X, self.target)
        print("Final Accuracy: %s"
              % accuracy_score(self.target, final_tfidf.predict(X_test)))

        # Final Accuracy: 0.882

    def baselineSVM(self):

        baseline_vectorizer = CountVectorizer(binary=True)
        baseline_vectorizer.fit(self.reviews_train_clean)
        X = baseline_vectorizer.transform(self.reviews_train_clean)
        X_test = baseline_vectorizer.transform(self.reviews_test_clean)

        X_train, X_val, y_train, y_val = train_test_split(
            X, self.target, train_size=0.75)

        for c in [0.01, 0.05, 0.25, 0.5, 1]:
            svm = LinearSVC(C=c)
            svm.fit(X_train, y_train)
            print("Accuracy for C=%s: %s"
                  % (c, accuracy_score(y_val, svm.predict(X_val))))

        final_svm_ngram = LinearSVC(C=0.01)
        final_svm_ngram.fit(X, self.target)
        print("Final Accuracy: %s"
              % accuracy_score(self.target, final_svm_ngram.predict(X_test)))

        # Final Accuracy: 0.8974
        self.print(baseline_vectorizer,final_svm_ngram)

    def print(self, vectorizer=None, final=None):
        feature_to_coef = {
            word: coef for word, coef in zip(
                vectorizer.get_feature_names(), final.coef_[0])
        }

        print("Best Positive")
        for best_positive in sorted(feature_to_coef.items(),
                                    key = lambda x: x[1],
                                    reverse = True)[: 30]:
            print(best_positive)

        print("\n\n")
        print("Best Negative")
        for best_negative in sorted(feature_to_coef.items(),
                                    key = lambda x: x[1])[: 30]:
            print(best_negative)

    def evaluate_model(self, model, test_pos_vec, test_neg_vec, print_confusion=False):
        """
        Prints the confusion matrix and accuracy of the model.
        """
        # Use the predict function and calculate the true/false positives and true/false negative.
        # YOUR CODE HERE
        predict1=model.predict(test_pos_vec).tolist()
        predict2=model.predict(test_neg_vec).tolist()
        #print predict1[0:5]
        #print type(predict1) is list
        #sys.exit(0)
        tp=predict1.count('pos')
        fn=predict1.count('neg')
        tn=predict2.count('neg')
        fp=predict2.count('pos')
        accuracy=float((tp+tn))/(len(test_pos_vec)+len(test_neg_vec))

        if print_confusion:
            print("predicted:\tpos\tneg")
            print("actual:")
            print("pos\t\t%d\t%d" % (tp, fn))
            print("neg\t\t%d\t%d" % (fp, tn))
        print("accuracy: %f" % (accuracy))

    def save(self, model, model_name='sentiment.model'):
        dump(model, model_name)

    def saveVectorizer(self, vectorizer, model_name='senti_vectorizer.model'):
        dump(vectorizer, model_name)

    def load(self, model_name):
        return load(model_name)

    def loadVectorizer(self, model_name='senti_vectorizer.model'):
        return load(model_name)

    def predict_sample(self, model, vectorizer_model):
        docs = ["한국 경제 글로벌 위기 수요 위축 시장 경제 붕귀 자동차 수출 빨간불 내수 촉진 증진 방향성 제고",
                "밝기 5등급 정도 도심 밖 맨눈 충분히 관측 가능 새해 미국인 8월 행운 기대",
                "최순실 민간인 국정농단 의혹 사건 진상규명 국정조사 특별위원회 1차 청문회 이재용 삼성전자 부회장 재벌 총수 9명 증인 출석"]

        text_features = vectorizer_model.transform(docs)
        predictions = model.predict(text_features)
        for text, predicted in zip(docs, predictions):
            print('"{}"'.format(text))
            print("  - Predicted as: '{}'".format(str(predicted)))
            print("")

    def predict(self, docs, model, vectorizer_model):
        text_features = vectorizer_model.transform(docs)
        predictions = model.predict(text_features)
        for text, predicted in zip(docs, predictions):
            print('"{}"'.format(text))
            print("  - Predicted as: '{}'".format(str(predicted)))
            print("")

        return predictions


''' tuple => tuple '''

class BaseTagger:
    IN_TYPE = [tuple]
    OUT_TYPE = [tuple]

class WordNet(BaseTagger):
    def __init__(self):
        import nltk
        nltk.download('wordnet')

        from nltk.stem import WordNetLemmatizer
        self.inst = WordNetLemmatizer()

    @staticmethod
    def get_wordnet_pos(treebank_tag):
        from nltk.corpus import wordnet
        if treebank_tag.startswith('J'):
            return wordnet.ADJ
        elif treebank_tag.startswith('V'):
            return wordnet.VERB
        elif treebank_tag.startswith('N'):
            return wordnet.NOUN
        elif treebank_tag.startswith('R'):
            return wordnet.ADV
        else:
            return None

    def __call__(self, *args, **kwargs):
        tag = WordNet.get_wordnet_pos(args[0][1])
        return self.inst.lemmatize(args[0][0], tag if tag else 'n'), args[0][1]


class KoreanLemmatizer:
    IN_TYPE = [str]
    OUT_TYPE = [list,str]

    def __init__(self):
        import soynlp
        from soynlp.lemmatizer import Lemmatizer
        from soynlp.lemmatizer import lemma_candidate
        test_stems = {
            '깨닫', '불', '묻', '눋', '겯', '믿', '묻', '뜯', '받',  # ㄷ 불규칙
            '구르', '무르', '마르', '누르', '나르', '모르', '이르',  # 르 불규칙
            '아니꼽', '우습', '더럽', '아름답', '잡', '뽑', '곱', '돕', '새롭', '더럽',  # ㅂ 불규칙
            '낫', '긋', '붓', '뭇', '벗', '솟', '치솟', '씻', '손씻', '뺏',  # ㅅ 불규칙
            '똥푸', '주', '좀주', '푸',  # 우 불규칙
            '끄', '크', '트', '모으',  # ㅡ 탈락 불규칙
            '삼가', '가', '들어가',  # 거라 불규칙
            '돌아오', '오',  # 너라 불규칙
            '이르', '푸르', '누르',  # 러 불규칙
            '하',  # 여 불규칙
            '가', '노랗', '퍼렇', '놀라',  # 어미 ㄴ
            '시퍼렇', '파랗',  # ㅎ 불규칙
            '먹', '먹이',
            '보', '뵈', '뵙', '그렇',
            '좋아지', '이',  # 이었 -> 였
            '만지',  # 지 -> 져
            '서툴', '내려놓',
        }

        test_eomis = {
            '',
            '아', '어나다', '어', '워', '웠다', '워서', '왔다', '와주니', '었다', '었어', '았어', '데',
            '라', '라니까', '너라', '았다', '러', '였다', '았다', '면', '다', '거라', '고', '는', '니', '었던', '엇어', '어서',
            'ㄴ', 'ㅂ고', '운', '았다'
        }

        self.inst = Lemmatizer(stems=test_stems, endings=test_eomis)

    def __call__(self, *args, **kwargs):
        #lemmatized=[]
        #for word in args[0]:
        #    lemmatized.append(self.inst.lemmatize(word))
        return self.inst.lemmatize(args[0])

class SejongPOSLemmatizer:
    IN_TYPE = [list, tuple]
    OUT_TYPE = [list, tuple]

    def __init__(self):
        from soylemma import Lemmatizer
        self.lemmatizer = Lemmatizer(dictionary_name='default')

    def __call__(self, *args, **kwargs):
        inst = []
        for i, word_tuple in enumerate(args[0]):
            #print(str(word_tuple))
            word = word_tuple[0]
            pos = word_tuple[1]
            if str(pos).startswith('XSA+') or str(pos).startswith('XSV+'):
                lemmatized = self.lemmatizer.lemmatize(word)
                #print(str(lemmatized))
                if len(lemmatized) > 1:
                    if (lemmatized[0][0] is '히다' and lemmatized[1][0] is '한다') or (lemmatized[1][0] is '히다' and lemmatized[0][0] is '한다'):
                        if (i > 0):
                            if str(args[0][i-1][1]).startswith('N'):
                                pre_term = inst[len(inst)-1]
                                inst.remove(pre_term)
                                inst.append((pre_term[0]+lemmatized[1][0],pre_term[1]))

                    else:
                        if (i > 0):
                            if str(args[0][i-1][1]).startswith('N'):
                                pre_term = inst[len(inst)-1]
                                inst.remove(pre_term)
                                inst.append((pre_term[0]+lemmatized[0][0],pre_term[1]))

                elif len(lemmatized) == 1:
                    if (i > 0):
                        if str(args[0][i - 1][1]).startswith('N'):
                            pre_term = inst[len(inst) - 1]

                            inst.remove(pre_term)
                            inst.append((pre_term[0] + lemmatized[0][0], pre_term[1]))
            elif str(pos) == 'XSA' or str(pos) == 'XSV':
                lemmatized = self.lemmatizer.lemmatize(word)
                if len(lemmatized) > 1:
                    if (lemmatized[0][0] is '히다' and lemmatized[1][0] is '한다') or (lemmatized[1][0] is '히다' and lemmatized[0][0] is '한다'):
                        if (i > 0):
                            if str(args[0][i-1][1]).startswith('N'):
                                pre_term = inst[len(inst)-1]
                                inst.remove(pre_term)
                                inst.append((pre_term[0]+lemmatized[1][0],pre_term[1]))

                    else:
                        if (i > 0):
                            if str(args[0][i-1][1]).startswith('N'):
                                pre_term = inst[len(inst)-1]
                                inst.remove(pre_term)
                                inst.append((pre_term[0]+lemmatized[0][0],pre_term[1]))
                elif len(lemmatized) == 1:
                    if (i > 0):
                        if str(args[0][i - 1][1]).startswith('N'):
                            pre_term = inst[len(inst) - 1]
                            inst.remove(pre_term)
                            inst.append((pre_term[0] + lemmatized[0][0], pre_term[1]))
            elif str(pos).startswith('VV+') or str(pos).startswith('VA+'):
                lemmatized = self.lemmatizer.lemmatize(word)
                #(str(lemmatized))
                if len(lemmatized) > 1:
                    if (lemmatized[0][0] is '히다' and lemmatized[1][0] is '한다') or (lemmatized[1][0] is '히다' and lemmatized[0][0] is '한다'):
                        if (i > 0):
                            if str(args[0][i-1][1]).startswith('N'):
                                pre_term = inst[len(inst) - 1]
                                inst.remove(pre_term)
                                inst.append((pre_term[0]+lemmatized[1][0],pre_term[1]))
                        else:
                            inst.append((lemmatized[0][0],pos))
                    else:
                        if (i > 0):
                            if str(args[0][i-1][1]).startswith('N'):
                                pre_term = inst[len(inst)-1]
                                inst.remove(pre_term)
                                inst.append((pre_term[0]+lemmatized[0][0],pre_term[1]))
                            elif str(pos).endswith('EC'):
                                if lemmatized[0][1] == 'Verb':
                                    inst.append((lemmatized[0][0], 'VV'))
                                elif lemmatized[1][1] == 'Verb':
                                    inst.append((lemmatized[1][0], 'VV'))
                        else:
                            inst.append((lemmatized[0][0],pos))
                elif len(lemmatized) == 1:
                    if (i > 0):
                        if str(args[0][i - 1][1]).startswith('N'):
                            pre_term = inst[len(inst) - 1]
                            inst.remove(pre_term)
                            inst.append((pre_term[0] + lemmatized[0][0], pre_term[1]))
                        else:
                            inst.append((lemmatized[0][0], pos))
                    else:
                        inst.append((lemmatized[0][0], pos))
            elif str(pos).startswith('VV') or str(pos).startswith('VA'):
                if (i > 0):
                    if str(args[0][i - 1][1]) == 'VX' or str(args[0][i - 1][1]) ==  'MAG':
                        pre_term = args[0][i - 1]
                        new_term = ''
                        if pre_term[0] == '않':
                            new_term = '안'
                        else:
                            new_term = pre_term[0]

                        if new_term == '안':
                            lemmatized = new_term + ' ' + word+'다'
                            inst.append((lemmatized, pre_term[1]))
                        else:
                            lemmatized = new_term + word + '다'
                            inst.append((lemmatized, pre_term[1]))
                    elif i > 0 or i < len(args[0]):
                        lemmatized = word + '다'
                        inst.append((lemmatized, pos))
                else:
                    lemmatized = word + '다'
                    if len(lemmatized) > 0:
                        inst.append((lemmatized,pos))

            elif str(pos).startswith('VX'):
                if i > 0 and i+1 < len(args[0]):
                    if str(args[0][i + 1][1]) == 'EC' and len(inst) > 0:
                        pre_term = inst[len(inst) - 1]
                        if (pre_term[1] == 'VA'):
                            inst.remove(pre_term)
                            lemmatized = '안 ' + pre_term[0]
                            inst.append((lemmatized, pos))
                        else:
                            lemmatized = word + '다'
                            inst.append((lemmatized, pos))


            elif (pos != 'EC' and pos != 'JX'):
                inst.append((word,pos))

        #print(inst)
        return inst



import math
import numpy as np


def tf_rf(corpus,package):
	labelset = package ["labelset"]
	weights = package ["weights"]
	doclist = package ["doclist"]
	doclist = dict(doclist)

	doclen = {}
	dictlist = {}
	worddict = {}
	totaldoc = len(corpus)

	for i in corpus:
		labell = i["label"]
		docl = i["document"]
		doclen[docl] = i["length"]
		if labell not in doclen:
			doclen[labell] = {}
		if docl not in doclen[labell]:
			doclen[labell][docl] = 0
		doclen[labell][docl] += i["length"]

		for j in i["split_sentence"].split():
			# dictlist : label —— doc —— word —— frequency
			if labell not in dictlist:
				dictlist[labell] = {}
			if docl not in dictlist[labell]:
				dictlist[labell][docl] = {}
			if j not in dictlist[labell][docl]:
				dictlist[labell][docl][j] = 0
			dictlist[labell][docl][j] += 1

			# oclist : label —— word ——　doc set
			if labell not in doclist:
				doclist[labell] = {}
			if j not in doclist[labell]:
				doclist[labell][j] = set()
			doclist[labell][j].add(docl)

	for labell in labelset:
		weights[labell] = {}
		for word in doclist[labell]:
			if word not in worddict:
				worddict[word] = 0
			worddict[word] += len(doclist[labell][word])
			a = len(doclist[labell][word])
			c = sum([len(doclist[labell][x]) for x in (doclist[labell]) if x!=word])
			weights[labell][word] = math.log(2+a*1.0/max(1,c*1.0),2)

	tf_rf = {}
	for labell in labelset:
		tf_rf[labell] = {}
		for doc in dictlist[labell]:
			tf_rf[labell][doc] = {}
			for word in dictlist[labell][doc]:
				#print doc + word
				tf_rf[labell][doc][word] = dictlist[labell][doc][word]*1.0 / (doclen[labell][doc]*1.0)
				if word in weights[labell]:
					tf_rf[labell][doc][word] *= weights[labell][word]

	package ["labelset"] = labelset
	package ["weights"] = weights
	package ["doclist"] = doclist
	return tf_rf



from collections import defaultdict

from sklearn.datasets import fetch_20newsgroups
from sklearn.decomposition import TruncatedSVD
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer
from sklearn.feature_extraction.text import HashingVectorizer
from sklearn.feature_extraction.text import TfidfTransformer
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import Normalizer
from sklearn import metrics
import matplotlib.pyplot as plt
from sklearn.cluster import KMeans, AgglomerativeClustering, SpectralClustering, SpectralCoclustering
import sys
from sklearn.decomposition import PCA
from scipy.sparse import csr_matrix
from time import time
import numpy as np
from numpy.linalg import svd
import operator

class DocumentClustering:
    def __init__(self, k=5):
        self.name = 'k-means'
        self.k = k
        self.X = None
        self.clustering = None
        self.vectorizer = None
        self.dataset_size=0
        self.doc2vec_matrix = False

    def make_matrix(self, documents=None, n_components=-1, doc2vec_matrix=None):
        if isinstance(doc2vec_matrix, np.ndarray) == False:
            self.vectorizer = TfidfVectorizer()
            # self.vectorizer = CountVectorizer()
            self.X = self.vectorizer.fit_transform(documents)
            self.dataset_size = len(documents)
        else:
            self.X = doc2vec_matrix
            self.dataset_size = len(doc2vec_matrix)
            self.doc2vec_matrix = True

        if (n_components != -1):
            if n_components > len(self.vectorizer.get_feature_names_out()):
                n_components = len(self.vectorizer.get_feature_names_out())
            print('n_components ' + str(n_components))
            # Vectorizer results are normalized, which makes KMeans behave as
            # spherical k-means for better results. Since LSA/SVD results are
            # not normalized, we have to redo the normalization.
            print("Performing dimensionality reduction using LSA")
            t0 = time()
            svd = TruncatedSVD(n_components)
            normalizer = Normalizer(copy=False)
            lsa = make_pipeline(svd, normalizer)

            self.X = lsa.fit_transform(self.X)

            print("done in %fs" % (time() - t0))

            explained_variance = svd.explained_variance_ratio_.sum()
            print("Explained variance of the SVD step: {}%".format(
                int(explained_variance * 100)))

            print()

    def cluster(self, cluster_name):
        self.name = cluster_name.strip()
        print('cluster_name ' + self.name)
        if self.name == 'k-means':
            print('cluster_name: ' + self.name)
            self.clustering = KMeans(n_clusters=self.k, init='k-means++', max_iter=500, n_init=1)
            print("Clustering sparse data with %s" % self.clustering)
            t0 = time()
            self.clustering.fit(self.X)
            print("done in %0.3fs" % (time() - t0))
            print()
        elif cluster_name == 'agglo':
            self.clustering = AgglomerativeClustering(n_clusters=self.k, affinity='euclidean', memory=None,
                                                      connectivity=None,
                                                      compute_full_tree='auto',
                                                      linkage='ward',
                                                      distance_threshold=None)

            print("Clustering sparse data with %s" % self.clustering)
            t0 = time()

            #to make dense matrix
            if self.doc2vec_matrix == False:
                self.X = self.X.tolist()
            self.clustering.fit(self.X)
            print("done in %0.3fs" % (time() - t0))
            print()
        elif self.name == 'spectral_cocluster':
            self.clustering = SpectralCoclustering(n_clusters=self.k,svd_method='arpack', random_state=0)
            print("Clustering sparse data with %s" % self.clustering)
            t0 = time()

            self.clustering.fit(self.X)
            print("done in %0.3fs" % (time() - t0))
            print()

    def print_results(self):
        # print the clustering result
        print(self.name)
        if self.name == 'k-means':
            cluster_labels = self.clustering.labels_
            clustering_dict = self.clustering.__dict__
            clusters = {}
            for document_id, cluster_label in enumerate(cluster_labels):
                if cluster_label not in clusters:
                    clusters[cluster_label] = []
                clusters[cluster_label].append(document_id)
                print(str(cluster_label) + " -- " + str(document_id))
            order_centroids = self.clustering.cluster_centers_.argsort()[:, ::-1]
            terms = self.vectorizer.get_feature_names_out()
            for i in range(self.k):
                print("Cluster %d:" % i, end='')
                for ind in order_centroids[i, :10]:
                    print(' %s' % terms[ind], end='')
                print()


        elif self.name == 'agglo':
            cluster_labels = self.clustering.labels_
            clustering_dict = self.clustering.__dict__
            clusters = {}

            for document_id, cluster_label in enumerate(cluster_labels):
                if cluster_label not in clusters:
                    clusters[cluster_label] = []
                clusters[cluster_label].append(document_id)
                #print(str(cluster_label) + " -- " + str(document_id))

            results = self.get_cluster_top_keywords(clusters)
            for _cluster in results:
                key_terms = results[_cluster]
                print("Cluster " + str(_cluster) + " : " + str(len(clusters[_cluster])) + " documents")
                print(key_terms)
            print()

        elif self.name == 'spectral_cocluster':
            target_number=10
            bicluster_ncuts = list(self.bicluster_ncut(i) for i in range(self.k))
            best_idx = np.argsort(bicluster_ncuts)[:target_number]

            feature_names = self.vectorizer.get_feature_names_out()
            print()
            print("Best biclusters:")
            print("----------------")
            for idx, cluster in enumerate(best_idx):
                n_rows, n_cols = self.clustering.get_shape(cluster)
                cluster_docs, cluster_words = self.clustering.get_indices(cluster)
                if not len(cluster_docs) or not len(cluster_words):
                    continue

                # categories
                counter = defaultdict(int)
                for i in cluster_docs:
                    counter[str(i)] += 1
                cat_string = ", ".join("{:.0f}% {}".format(float(c) / n_rows * 100, name) for name, c in self.most_common(counter)[:3])

                # words
                out_of_cluster_docs = self.clustering.row_labels_ != cluster
                out_of_cluster_docs = np.where(out_of_cluster_docs)[0]
                word_col = self.X[:, cluster_words]
                word_scores = np.array(word_col[cluster_docs, :].sum(axis=0) -
                                       word_col[out_of_cluster_docs, :].sum(axis=0))
                word_scores = word_scores.ravel()
                important_words = list(feature_names[cluster_words[i]]
                                       for i in word_scores.argsort()[:-11:-1])

                print("bicluster {} : {} documents, {} words".format(idx, n_rows, n_cols))
                print("categories   : {}".format(cat_string))
                print("words        : {}\n".format(', '.join(important_words)))

    def bicluster_ncut(self, i):
        rows, cols = self.clustering.get_indices(i)
        if not (np.any(rows) and np.any(cols)):
            import sys
            return sys.float_info.max

        row_complement = np.nonzero(np.logical_not(self.clustering.rows_[i]))[0]
        col_complement = np.nonzero(np.logical_not(self.clustering.columns_[i]))[0]
        # Note: the following is identical to X[rows[:, np.newaxis],
        # cols].sum() but much faster in scipy <= 0.16
        weight = self.X[rows][:, cols].sum()
        cut = (self.X[row_complement][:, cols].sum() + self.X[rows][:, col_complement].sum())

        return cut / weight

    def most_common(self, d):
        """Items of a defaultdict(int) with the highest values.
        """
        return sorted(d.items(), key=operator.itemgetter(1), reverse=True)

    def get_cluster_top_keywords(self, clusters, keywords_per_cluster=10):
        """Shows the top k words for each cluster
        Keyword Arguments:
            keywords_per_cluster {int} -- The k words to show for each cluster (default: {10})
        Returns:
            dict of lists -- Returns a dict of {cluster_id: ['top', 'k', 'words', 'for', 'cluster']}
        """
        terms = self.vectorizer.get_feature_names_out()
        out = {}
        docs_for_cluster = {}
        # self.clusters = 10 clusters,containing the index of the document_vectors document in that cluster, ex len(self.clusters[6]) == 508
        for cluster in clusters:
            # To flatten/combine all documents into one
            docs_for_cluster[cluster] = np.array([self.X[i] for i in clusters[cluster]])
            # Cluster vectors to feature words
            out[cluster] = np.array(terms)[np.flip(np.argsort(docs_for_cluster[cluster]), -1)]
            cluster_shape = out[cluster].shape
            out[cluster] = out[cluster].reshape(cluster_shape[0] * cluster_shape[1])[:keywords_per_cluster].tolist()

        return out

    def visualize(self):
        # The output is a one-dimensional array of N documents corresponding to the clusters
        # assigned to our N data points.
        if self.name == 'spectral_cocluster':
            pca_t = None
            if self.doc2vec_matrix == False:
                if type(self.X) is np.ndarray:
                    pca_t = PCA().fit_transform(self.X.tolist())
                else:
                    pca_t = PCA().fit_transform(self.X.toarray())
            else:
                pca_t = PCA().fit_transform(self.X)
            #pca_t = PCA().fit_transform(self.X)
            # print(self.clustering.labels_)
            plt.scatter(pca_t[:, 0], pca_t[:, 1], c=self.clustering.row_labels_, cmap='rainbow')
            plt.show()
        elif self.name == 'agglo':
            pca_t = PCA().fit_transform(self.X)
            plt.scatter(pca_t[:, 0], pca_t[:, 1], c=self.clustering.labels_, cmap='rainbow')
            plt.show()
        elif self.name == 'k-means':
            if self.doc2vec_matrix == False and not isinstance(self.X,np.ndarray):
                self.X = self.X.toarray()

            pca_t = PCA().fit_transform(self.X)
            # print(self.clustering.labels_)
            plt.scatter(pca_t[:, 0], pca_t[:, 1], c=self.clustering.labels_, cmap='rainbow')
            plt.show()


from itertools import chain

class BaseNgram:
    IN_TYPE = [list, str]
    OUT_TYPE = [list, str]

class NGramTokenizer(BaseNgram):
    def __init__(self, min=1, ngramCount=3, concat='_'):
        self.ngramCount = ngramCount
        self.min = min
        self.converted = []
        self.concat = concat

    def __call__(self, *args, **kwargs):
        converted = []
        from nltk.util import ngrams
        for i in range(self.min, self.ngramCount+1):
            output = list(ngrams((args[0]), i))
            for x in output:
                if (len(x) > 0):
                    converted.append(self.concat.join(x))

        #print("NGRAM " + str(converted))
        self.converted = converted

        return converted


from pprint import pprint
from pycrfsuite_spacing import TemplateGenerator
from pycrfsuite_spacing import CharacterFeatureTransformer
from pycrfsuite_spacing import sent_to_chartags
from pycrfsuite_spacing import sent_to_xy
from pycrfsuite_spacing import PyCRFSuiteSpacing

with open('../../data/134963_norm.txt', encoding='utf-8') as f:
    docs = [doc.strip() for doc in f]

print('n docs = %d' % len(docs))
pprint(docs[:5])

to_feature = CharacterFeatureTransformer(
    TemplateGenerator(begin=-2,
    end=2,
    min_range_length=3,
    max_range_length=3)
)

x, y = sent_to_xy('이것도 너프해 보시지', to_feature)
pprint(x)
print(y)

correct = PyCRFSuiteSpacing(
    to_feature = to_feature,
    feature_minfreq=3, # default = 0
    max_iterations=100,
    l1_cost=1.0,
    l2_cost=1.0
)
correct.train(docs, '../../model/korean_segmentation_model.crfsuite')




from treform.collector.colBasic import *

class ColNaverNews(ColBasic):
    def getInfo(params):
        return '네이버 뉴스 수집기', 0

    def getName(self):
        return "colNaverNews"

    def getSearchURL(self):
        return 'https://m.search.naver.com/search.naver'

    def makeParameter(self, page, query, d):
        if page*16 >= 4000: raise Exception("Cannot collect over 4000 articles")
        return {'where':'m_news',
                'query':query,
                'sm':'mtb_tnw',
                'sort':1,
                'pd':3,
                'ds':self.reformatDate(d['startDate'], '%04d.%02d.%02d'),
                'de':self.reformatDate(d['endDate'], '%04d.%02d.%02d'),
                'start':page*16 - 15,
            }

    def selectList(self, soup):
        ret = []
        for el in soup.select('.list_news a'):
            link = el['href']
            if link.find('news.naver.com') < 0: continue
            ret.append(link)
        return ret

    def selectArticle(self, soup):
        [x.extract() for x in soup.findAll('script')]
        return (ColNaverNews.cleanText(soup.select_one('.media_end_head_info_datestamp_time').text),
                ColNaverNews.cleanText(soup.select_one('h2').text),
                ColNaverNews.cleanText(soup.select_one('.newsct_article').text))



import networkx as nx
import matplotlib.pyplot as plt
import numpy as np
import matplotlib.font_manager as fm
import platform
from matplotlib.ft2font import FT2Font
import matplotlib as mpl

class GraphMLCreator:

    def __init__(self):
        self.G = nx.Graph()

        # Hack: offset the most central node to avoid too much overlap
        self.rad0 = 0.3

    def createGraphML(self, co_occurrence, vocabulary, file):
        G = nx.Graph()

        for obj in vocabulary:
            G.add_node(obj)
        # convert list to a single dictionary

        for pair in co_occurrence:
            node1 = ''
            node2 = ''
            #for inner_pair in pair:

            if type(pair) is tuple:
                node1 = pair[0]
                node2 = pair[1]
            elif type(pair) is str:
                pair=pair.split()
                node1 = pair[0]
                node2 = pair[1]
            elif type(pair) is int:
                #print ("X " + node1 + " == " + node2 + " == " + str(inner_pair) + " : " + str(tuple[node1]))
                G.add_edge(node1, node2, weight=float(pair))
            elif type(pair) is float:
                #print ("X " + node1 + " == " + node2 + " == " + str(inner_pair) + " : ")
                G.add_edge(node1, node2, weight=float(pair))

        self.G = G
        print(self.G.number_of_nodes())
        nx.write_graphml(G, file)

    def createGraphMLWithThreshold(self, co_occurrence, word_hist, vocab, file, threshold=10.0):
        G = nx.Graph()

        filtered_word_list=[]
        for pair in co_occurrence:
            node1 = ''
            node2 = ''
            for inner_pair in pair:
                if type(inner_pair) is tuple:
                    node1 = inner_pair[0]
                    node2 = inner_pair[1]
                elif type(inner_pair) is str:
                    inner_pair=inner_pair.split()
                    node1 = inner_pair[0]
                    node2 = inner_pair[1]
                elif type(inner_pair) is int:
                    if float(inner_pair) >= threshold:
                        #print ("X " + node1 + " == " + node2 + " == " + str(inner_pair) + " : " + str(tuple[node1]))
                        G.add_edge(node1, node2, weight=float(inner_pair))
                        if node1 not in filtered_word_list:
                            filtered_word_list.append(node1)
                        if node2 not in filtered_word_list:
                            filtered_word_list.append(node2)
                elif type(inner_pair) is float:
                    if float(inner_pair) >= threshold:
                        #print ("X " + node1 + " == " + node2 + " == " + str(inner_pair) + " : ")
                        G.add_edge(node1, node2, weight=float(inner_pair))
                        if node1 not in filtered_word_list:
                            filtered_word_list.append(node1)
                        if node2 not in filtered_word_list:
                            filtered_word_list.append(node2)
        for word in word_hist:
            if word in filtered_word_list:
                G.add_node(word, count=word_hist[word])

        self.G = G
        print(self.G.number_of_nodes())
        nx.write_graphml(G, file)

    def createGraphMLWithThresholdInDictionary(self, co_occurrence, word_hist, file, threshold=10.0):
        G = nx.Graph()

        node1 = ''
        node2 = ''
        filtered_word_list=[]
        for pair in co_occurrence:
            node1 = str(pair[0])
            node2 = str(pair[1])

            inner_pair = co_occurrence[pair]

            if type(inner_pair) is str:
                inner_pair=inner_pair.split()
                node1 = inner_pair[0]
                node2 = inner_pair[1]
            elif type(inner_pair) is int:
                if float(inner_pair) >= threshold:
                    #print ("X " + node1 + " == " + node2 + " == " + str(inner_pair) + " : " + str(tuple[node1]))
                    G.add_edge(node1, node2, weight=float(inner_pair))
                    if node1 not in filtered_word_list:
                        filtered_word_list.append(node1)
                    if node2 not in filtered_word_list:
                        filtered_word_list.append(node2)
            elif type(inner_pair) is float:
                if float(inner_pair) >= threshold:
                    #print ("X " + node1 + " == " + node2 + " == " + str(inner_pair) + " : ")
                    G.add_edge(node1, node2, weight=float(inner_pair))
                    if node1 not in filtered_word_list:
                        filtered_word_list.append(node1)
                    if node2 not in filtered_word_list:
                        filtered_word_list.append(node2)

        for word in word_hist:
            if word in filtered_word_list:
                G.add_node(word, count=word_hist[word])

        self.G = G
        print(self.G.number_of_nodes())
        nx.write_graphml(G, file)

    def centrality_layout(self):
        centrality = nx.eigenvector_centrality_numpy(self.G)
        """Compute a layout based on centrality.
        """
        # Create a list of centralities, sorted by centrality value
        cent = sorted(centrality.items(), key=lambda x:float(x[1]), reverse=True)
        nodes = [c[0] for c in cent]
        cent  = np.array([float(c[1]) for c in cent])
        rad = (cent - cent[0])/(cent[-1]-cent[0])
        rad = self.rescale_arr(rad, self.rad0, 1)
        angles = np.linspace(0, 2*np.pi, len(centrality))
        layout = {}
        for n, node in enumerate(nodes):
            r = rad[n]
            th = angles[n]
            layout[node] = r*np.cos(th), r*np.sin(th)
        return layout

    def plot_graph(self, title=None, file='graph.png'):
        #from matplotlib.font_manager import _rebuild
        #_rebuild()

        font_path = ''
        if platform.system() is 'Windows':
            # Window의 경우 폰트 경로
            font_path = 'C:/Windows/Fonts/malgun.ttf'
        elif platform.system() is 'Darwin':
            # for Mac
            font_path='/Library/Fonts/AppleGothic.ttf'

        font_name = fm.FontProperties(fname=font_path).get_name()
        plt.rc('font', family=font_name)
        plt.rc('axes', unicode_minus=False)
        # 그래프에서 마이너스 폰트 깨지는 문제에 대한 대처
        mpl.rcParams['axes.unicode_minus'] = False
        #print('버전: ', mpl.__version__)
        #print('설치 위치: ', mpl.__file__)
        #print('설정 위치: ', mpl.get_configdir())
        #print('캐시 위치: ', mpl.get_cachedir())

        # size, family
        print('# 설정 되어있는 폰트 사이즈')
        print(plt.rcParams['font.size'])
        print('# 설정 되어있는 폰트 글꼴')
        print(plt.rcParams['font.family'])

        fig = plt.figure(figsize=(8, 8))
        pos = self.centrality_layout()

        """Conveniently summarize graph visually"""
        # config parameters
        edge_min_width= 3
        edge_max_width= 12
        label_font = 18
        node_font = 22
        node_alpha = 0.4
        edge_alpha = 0.55
        edge_cmap = plt.cm.Spectral

        # Create figure
        if fig is None:
            fig, ax = plt.subplots()
        else:
            ax = fig.add_subplot(111)
        fig.subplots_adjust(0,0,1)

        font = FT2Font(font_path)

        # Plot nodes with size according to count
        sizes = []
        degrees = []
        for n, d in self.G.nodes(data=True):
            sizes.append(d['count'])
            degrees.append(self.G.degree(n))

        sizes = self.rescale_arr(np.array(sizes, dtype=float), 100, 1000)

        # Compute layout and label edges according to weight
        pos = nx.spring_layout(self.G) if pos is None else pos
        labels = {}
        width = []
        for n1, n2, d in self.G.edges(data=True):
            w = d['weight']
            labels[n1, n2] = w
            width.append(w)

        width = self.rescale_arr(np.array(width, dtype=float), edge_min_width,
                            edge_max_width)

        # Draw
        nx.draw_networkx_nodes(self.G, pos, node_size=sizes, node_color=degrees,
                               alpha=node_alpha)
        nx.draw_networkx_edges(self.G, pos, width=width, edge_color=width,
                               edge_cmap=edge_cmap, alpha=edge_alpha)
        nx.draw_networkx_edge_labels(self.G, pos, edge_labels=labels,
                                     font_size=label_font)
        nx.draw_networkx_labels(self.G, pos, font_size=node_font, font_family=font_name, font_weight='bold')

        if title is not None:
            ax.set_title(title, fontsize=label_font)
        ax.set_xticks([])
        ax.set_yticks([])

        # Mark centrality axes
        kw = dict(color='k', linestyle='-')
        cross = [ax.axhline(0, **kw), ax.axvline(self.rad0, **kw)]
        [l.set_zorder(0) for l in cross]

        #plt.show()
        plt.savefig(file)

    def rescale_arr(self, arr, amin, amax):
        """Rescale an array to a new range.
        Return a new array whose range of values is (amin, amax).
        Parameters
        ----------
        arr : array-like
        amin : float
          new minimum value
        amax : float
          new maximum value
        Examples
        --------
        #>>> a = np.arange(5)
        #>>> rescale_arr(a,3,6)
        array([ 3.  ,  3.75,  4.5 ,  5.25,  6.  ])
        """

        # old bounds
        m = arr.min()
        M = arr.max()
        # scale/offset
        s = float(amax - amin) / (M - m)
        d = amin - s * m

        # Apply clip before returning to cut off possible overflows outside the
        # intended range due to roundoff error, so that we can absolutely guarantee
        # that on output, there are no values > amax or < amin.
        return np.clip(s * arr + d, amin, amax)

    def summarize_centrality(self, limit=10):

        centrality = nx.eigenvector_centrality_numpy(self.G)
        c = centrality.items()
        c = sorted(c, key=lambda x: x[1], reverse=True)
        print('\nGraph centrality')
        count=0
        for node, cent in c:
            if count>limit:
                break
            print ("%15s: %.3g" % (node, float(cent)))
            count+=1

    def sort_freqs(self, freqs):
        """Sort a word frequency histogram represented as a dictionary.
        Parameters
        ----------
        freqs : dict
          A dict with string keys and integer values.
        Return
        ------
        items : list
          A list of (count, word) pairs.
        """
        items = freqs.items()
        items.sort(key=lambda wc: wc[1])
        return items

    def plot_word_histogram(self, freqs, show=10, title=None):
        """Plot a histogram of word frequencies, limited to the top `show` ones.
        """
        sorted_f = self.sort_freqs(freqs) if isinstance(freqs, dict) else freqs

        # Don't show the tail
        if isinstance(show, int):
            # interpret as number of words to show in histogram
            show_f = sorted_f[-show:]
        else:
            # interpret as a fraction
            start = -int(round(show * len(freqs)))
            show_f = sorted_f[start:]

        # Now, extract words and counts, plot
        n_words = len(show_f)
        ind = np.arange(n_words)
        words = [i[0] for i in show_f]
        counts = [i[1] for i in show_f]

        fig = plt.figure()
        ax = fig.add_subplot(111)

        if n_words <= 20:
            # Only show bars and x labels for small histograms, they don't make
            # sense otherwise
            ax.bar(ind, counts)
            ax.set_xticks(ind)
            ax.set_xticklabels(words, rotation=45)
            fig.subplots_adjust(bottom=0.25)
        else:
            # For larger ones, do a step plot
            ax.step(ind, counts)

        # If it spans more than two decades, use a log scale
        if float(max(counts)) / min(counts) > 100:
            ax.set_yscale('log')

        if title:
            ax.set_title(title)
        return ax


import setuptools
from treform import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

#torch for CPU only
#pip install torch==1.5.1+cpu torchvision==0.6.1+cpu -f https://download.pytorch.org/whl/torch_stable.html

setuptools.setup(
    name="treform_new", # Replace with your own username
    version=__version__,
    author="Min Song",
    author_email="min.song@yonsei.ac.kr",
    description="A text mining tool for Korean and English",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MinSong2/treform",
    packages=setuptools.find_packages(exclude = []),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "cython",
        "seqeval",
        "pytorch-crf",
        "numpy",
        "kiwipiepy",
        "transformers",
        "sklearn-crfsuite",
        "gensim",
        "konlpy",
        "krwordrank",
        "lxml",
        "matplotlib",
        "networkx",
        "node2vec",
        "bs4",
        "pycrfsuite-spacing",
        "scikit-learn",
        "scipy==1.10.1",
        "seaborn",
        "soynlp",
        "torch",
        "tomotopy",
        "pyLDAvis",
        "wordcloud",
        "nltk",
        "newspaper3k",
        "selenium",
        "soylemma",
        "bokeh",
        "beautifulsoup4",
        "benepar>=0.2.0",
        "symspellpy"
    ],
    python_requires='>=3.10',
)


import newspaper3k
import os
import time
import random
import sys
import re

reSpace = re.compile('\\s+')
metaDatePublished = re.compile('<meta *itemprop="datePublished" *content="([^"]+)" */?>')
metaUploadDate = re.compile('<meta *itemprop="uploadDate" *content="([^"]+)" */?>')
metaUtime = re.compile('<meta *name="utime" *content="([0-9]{4})([0-9]{2})([0-9]{2}).*?" */?>')
metaPdate = re.compile('<meta *name="pdate" *content="([0-9]{4})([0-9]{2})([0-9]{2}).*?" */?>')
rawDateText = re.compile('(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)[^\\s]* *([0-9]{1,2}), *([0-9]{4})')

def extractDate(html):
    m = metaDatePublished.search(html)
    if m: return m.group(1)
    m = metaUploadDate.search(html)
    if m: return m.group(1)
    m = metaUtime.search(html)
    if m: return m.group(1) + '-' + m.group(2) + '-' + m.group(3)
    m = metaPdate.search(html)
    if m: return m.group(1) + '-' + m.group(2) + '-' + m.group(3)
    m = rawDateText.search(html)
    if m: return m.group(0)
    return None

def getArticle(url):
    import dateutil.parser
    article = newspaper.Article(url, keep_article_html=True)
    try:
        article.download()
        article.parse()
    except Exception as e:
        raise e
    date = article.publish_date or dateutil.parser.parse(extractDate(article.html)).strftime('%Y-%m-%d %H:%M:%S')
    if not date: raise Exception("Cannot find date")
    return (url, article.title, ','.join(article.authors), str(date), article.text.replace('\n', ' '))

if __name__ == "__main__":
    keyword = sys.argv[1]
    output = open('data/contents/%s.txt' % keyword, 'w', encoding='utf-8')
    for name in os.listdir('data/url'):
        if name.find(keyword) < 0: continue
        for url in set(open('data/url/' + name, encoding='utf-8').readlines()):
            print("Download %s ..." % url.strip())
            try:
                output.write("\t".join(getArticle(url.strip())) + "\n")
            except:
                print("Failed to fetch page : %s"  % url.strip())
            time.sleep(random.randint(4, 10))
    output.close()


''' str => list(str) '''

class BaseSplitter:
    IN_TYPE = [str]
    OUT_TYPE = [list, str]

class SpecialCharRemover(BaseSplitter):
    IN_TYPE = [str]
    OUT_TYPE = [str]

    def __init__(self):
        import re
        self.hangul = re.compile('[^ ㄱ-ㅣ가-힣\\.\\?\\,]+')

    def __call__(self, *args, **kwargs):
        return self.hangul.sub('', *args)

class NLTK(BaseSplitter):
    def __init__(self):
        import nltk.tokenize
        self.func = nltk.tokenize.sent_tokenize

    def __call__(self, *args, **kwargs):
        return self.func(*args)


import logging
import regex
import sys
from collections import defaultdict, Counter
from treform.abbreviations.candidate import Candidate

"""
A Simple Algorithm for Identifying Abbreviations Definitions in Biomedical Text
A. Schwartz and M. Hearst
Biocomputing, 2003, pp 451-462.

"""

class BaseAbbreviationExtraction:
    IN_TYPE = [str]
    OUT_TYPE = [tuple]

class SchwartzHearstAbbreviationExtraction(BaseAbbreviationExtraction):
    def __init__(self):
        logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s', level=logging.INFO)
        self.log = logging.getLogger(__name__)

    def yield_lines_from_file(self, file_path):
        with open(file_path, 'rb') as f:
            for line in f:
                try:
                    line = line.decode('utf-8')
                except UnicodeDecodeError:
                    line = line.decode('latin-1').encode('utf-8').decode('utf-8')
                line = line.strip()
                yield line


    def yield_lines_from_doc(self, doc_text):
        for line in doc_text.split("\n"):
            yield line.strip()


    def best_candidates(self, sentence):
        """
        :param sentence: line read from input file
        :return: a Candidate iterator
        """

        if '(' in sentence:
            # Check some things first
            if sentence.count('(') != sentence.count(')'):
                raise ValueError("Unbalanced parentheses: {}".format(sentence))

            if sentence.find('(') > sentence.find(')'):
                raise ValueError("First parentheses is right: {}".format(sentence))

            close_index = -1
            while 1:
                # Look for open parenthesis. Need leading whitespace to avoid matching mathematical and chemical formulae
                open_index = sentence.find(' (', close_index + 1)

                if open_index == -1: break

                # Advance beyond whitespace
                open_index += 1

                # Look for closing parentheses
                close_index = open_index + 1
                open_count = 1
                skip = False
                while open_count:
                    try:
                        char = sentence[close_index]
                    except IndexError:
                        # We found an opening bracket but no associated closing bracket
                        # Skip the opening bracket
                        skip = True
                        break
                    if char == '(':
                        open_count += 1
                    elif char in [')', ';', ':']:
                        open_count -= 1
                    close_index += 1

                if skip:
                    close_index = open_index + 1
                    continue

                # Output if conditions are met
                start = open_index + 1
                stop = close_index - 1
                candidate = sentence[start:stop]

                # Take into account whitespace that should be removed
                start = start + len(candidate) - len(candidate.lstrip())
                stop = stop - len(candidate) + len(candidate.rstrip())
                candidate = sentence[start:stop]

                if self.conditions(candidate):
                    new_candidate = Candidate(candidate)
                    new_candidate.set_position(start, stop)
                    yield new_candidate


    def conditions(self, candidate):
        """
        Based on Schwartz&Hearst

        2 <= len(str) <= 10
        len(tokens) <= 2
        re.search(r'\p{L}', str)
        str[0].isalnum()

        and extra:
        if it matches (\p{L}\.?\s?){2,}
        it is a good candidate.

        :param candidate: candidate abbreviation
        :return: True if this is a good candidate
        """
        viable = True
        if regex.match(r'(\p{L}\.?\s?){2,}', candidate.lstrip()):
            viable = True
        if len(candidate) < 2 or len(candidate) > 10:
            viable = False
        if len(candidate.split()) > 2:
            viable = False
        if not regex.search(r'\p{L}', candidate):
            viable = False
        if not candidate[0].isalnum():
            viable = False

        return viable


    def get_definition(self, candidate, sentence):
        """
        Takes a candidate and a sentence and returns the definition candidate.

        The definition candidate is the set of tokens (in front of the candidate)
        that starts with a token starting with the first character of the candidate

        :param candidate: candidate abbreviation
        :param sentence: current sentence (single line from input file)
        :return: candidate definition for this abbreviation
        """
        # Take the tokens in front of the candidate
        tokens = regex.split(r'[\s\-]+', sentence[:candidate.start - 2].lower())
        # the char that we are looking for
        key = candidate[0].lower()

        # Count the number of tokens that start with the same character as the candidate
        first_chars = [t[0] for t in filter(None, tokens)]

        definition_freq = first_chars.count(key)
        candidate_freq = candidate.lower().count(key)

        # Look for the list of tokens in front of candidate that
        # have a sufficient number of tokens starting with key
        if candidate_freq <= definition_freq:
            # we should at least have a good number of starts
            count = 0
            start = 0
            start_index = len(first_chars) - 1
            while count < candidate_freq:
                if abs(start) > len(first_chars):
                    raise ValueError("candidate {} not found".format(candidate))
                start -= 1
                # Look up key in the definition
                try:
                    start_index = first_chars.index(key, len(first_chars) + start)
                except ValueError:
                    pass

                # Count the number of keys in definition
                count = first_chars[start_index:].count(key)

            # We found enough keys in the definition so return the definition as a definition candidate
            start = len(' '.join(tokens[:start_index]))
            stop = candidate.start - 1
            candidate = sentence[start:stop]

            # Remove whitespace
            start = start + len(candidate) - len(candidate.lstrip())
            stop = stop - len(candidate) + len(candidate.rstrip())
            candidate = sentence[start:stop]

            new_candidate = Candidate(candidate)
            new_candidate.set_position(start, stop)
            return new_candidate

        else:
            raise ValueError('There are less keys in the tokens in front of candidate than there are in the candidate')


    def select_definition(self, definition, abbrev):
        """
        Takes a definition candidate and an abbreviation candidate
        and returns True if the chars in the abbreviation occur in the definition

        Based on
        A simple algorithm for identifying abbreviation definitions in biomedical texts, Schwartz & Hearst
        :param definition: candidate definition
        :param abbrev: candidate abbreviation
        :return:
        """

        if len(definition) < len(abbrev):
            raise ValueError('Abbreviation is longer than definition')

        if abbrev in definition.split():
            raise ValueError('Abbreviation is full word of definition')

        s_index = -1
        l_index = -1

        while 1:
            try:
                long_char = definition[l_index].lower()
            except IndexError:
                raise

            short_char = abbrev[s_index].lower()

            if not short_char.isalnum():
                s_index -= 1

            if s_index == -1 * len(abbrev):
                if short_char == long_char:
                    if l_index == -1 * len(definition) or not definition[l_index - 1].isalnum():
                        break
                    else:
                        l_index -= 1
                else:
                    l_index -= 1
                    if l_index == -1 * (len(definition) + 1):
                        raise ValueError("definition {} was not found in {}".format(abbrev, definition))

            else:
                if short_char == long_char:
                    s_index -= 1
                    l_index -= 1
                else:
                    l_index -= 1

        new_candidate = Candidate(definition[l_index:len(definition)])
        new_candidate.set_position(definition.start, definition.stop)
        definition = new_candidate

        tokens = len(definition.split())
        length = len(abbrev)

        if tokens > min([length + 5, length * 2]):
            raise ValueError("did not meet min(|A|+5, |A|*2) constraint")

        # Do not return definitions that contain unbalanced parentheses
        if definition.count('(') != definition.count(')'):
            raise ValueError("Unbalanced parentheses not allowed in a definition")

        return definition


    def extract_abbreviation_definition_pairs(self,
                                              file_path=None,
                                              doc_text=None,
                                              most_common_definition=False,
                                              first_definition=False):
        abbrev_map = dict()
        list_abbrev_map = defaultdict(list)
        counter_abbrev_map = dict()
        omit = 0
        written = 0
        if file_path:
            sentence_iterator = enumerate(self.yield_lines_from_file(file_path))
        elif doc_text:
            sentence_iterator = enumerate(self.yield_lines_from_doc(doc_text))
        else:
            return abbrev_map

        collect_definitions = False
        if most_common_definition or first_definition:
            collect_definitions = True

        for i, sentence in sentence_iterator:
            # Remove any quotes around potential candidate terms
            clean_sentence = regex.sub(r'([(])[\'"\p{Pi}]|[\'"\p{Pf}]([);:])', r'\1\2', sentence)
            try:
                for candidate in self.best_candidates(clean_sentence):
                    try:
                        definition = self.get_definition(candidate, clean_sentence)
                    except (ValueError, IndexError) as e:
                        self.log.debug("{} Omitting candidate {}. Reason: {}".format(i, candidate, e.args[0]))
                        omit += 1
                    else:
                        try:
                            definition = self.select_definition(definition, candidate)
                        except (ValueError, IndexError) as e:
                            self.log.debug("{} Omitting definition {} for candidate {}. Reason: {}".format(i, definition, candidate, e.args[0]))
                            omit += 1
                        else:
                            # Either append the current definition to the list of previous definitions ...
                            if collect_definitions:
                                list_abbrev_map[candidate].append(definition)
                            else:
                                # Or update the abbreviations map with the current definition
                                abbrev_map[candidate] = definition
                            written += 1
            except (ValueError, IndexError) as e:
                self.log.debug("{} Error processing sentence {}: {}".format(i, sentence, e.args[0]))
        self.log.debug("{} abbreviations detected and kept ({} omitted)".format(written, omit))

        # Return most common definition for each term
        if collect_definitions:
            if most_common_definition:
                # Return the most common definition for each term
                for k,v in list_abbrev_map.items():
                    counter_abbrev_map[k] = Counter(v).most_common(1)[0][0]
            else:
                # Return the first definition for each term
                for k, v in list_abbrev_map.items():
                    counter_abbrev_map[k] = v[0]
            return counter_abbrev_map

        # Or return the last encountered definition for each term
        return abbrev_map

    def __call__(self, *args, **kwargs):
        pairs = self.extract_abbreviation_definition_pairs(doc_text=args[0], most_common_definition=True)

        return pairs




import numpy  # for arrays, array broadcasting etc.
from gensim.models import ldaseqmodel, ldamodel
from gensim.corpora import Dictionary
import os.path
import logging

module_path = os.path.dirname(__file__)  # needed because sample data files are located in the same folder

class ldaSeqModel():

    def __init__(self):
        name = 'ldaSeqModel'

    def run(self, document_collection, topic_count=2, time_group=[10,10,11]):
        """document_collection should be sorted in order of time_slice."""
        dictionary = Dictionary(document_collection)
        corpus = [dictionary.doc2bow(text) for text in document_collection]
        ldaseq = ldaseqmodel.LdaSeqModel(corpus=corpus, id2word=dictionary, num_topics=topic_count, time_slice=time_group)

        topics = ldaseq.print_topics(1)
        for topic in topics:
            print("TOPIC " + str(topic))

        return ldaseq

    def parseDocuments(self, document_file, year_index, document_index):
        """make document along with time information"""
        dict = {}
        with open(document_file, encoding='utf-8') as ins:
            for line in ins:
                #print("LINE " + line)
                fields = line.split('\t')
                _year = fields[year_index]
                _document = fields[document_index]

                if _year not in dict:
                    d = []
                    d.append(_document)
                    dict[_year] = d

                else:
                    print("DOC " + _year)
                    _docu_ = dict.get(_year)
                    _docu_.append(_document)
        return dict

    def parseProcessedText(self, processed_documents, pair_map):
        """make document along with time information"""
        dict = {}
        for doc in processed_documents:
            for line in doc:
                #print("LINE " + line)
                fields = line.split('\t')
                _year = fields[year_index]
                _document = fields[document_index]

                if _year not in dict:
                    d = []
                    d.append(_document)
                    dict[_year] = d

                else:
                    print("DOC " + _year)
                    _docu_ = dict.get(_year)
                    _docu_.append(_document)
        return dict

if __name__ == '__main__':
    logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s', level=logging.DEBUG)
    document_file = "../time_test.txt"
    year_index = 0
    document_index = 1
    _dict = ldaSeqModel().parseDocuments(document_file, year_index, document_index)

    #import pyTextMiner as ptm
    #corpus = ptm.CorpusFromFieldDelimitedFileWithYear('time_test.txt', 1, 0)
    #pair_map = corpus.pair_map

    time_slice = []
    key_size = len(_dict)
    doc_coll = _dict.values()
    for k, v in _dict.items():
        time_slice.append(len(v))
    ldaSeqModel().run(doc_coll,5,time_slice)



from treform.collector.colBasic import *

class ColChosun(ColBasic):
    def getInfo(params):
        return '조선일보 수집기', 0

    def getName(self):
        return "colChosun"

    def getSearchURL(self):
        return 'https://nsearch.chosun.com/search/total.search'

    def makeParameter(self, page, query, d):
        return {'query':query,
                    'pageno':page,
                    'orderby':'docdatetime',
                    'naviarraystr':'',
                    'kind':'',
                    'cont1':'',
                    'cont2':'',
                    'cont5':'',
                    'categoryname':'',
                    'categoryd2':'',
                    'c_scope':'paging',
                    'sdate':d['startDate'],
                    'edate':d['endDate'],
                    'premium':'',
            }

    def selectList(self, soup):
        ret = []
        for el in soup.select('.search_news dt a'):
            link = el['href']
            ret.append(link)
        return ret

    def selectArticle(self, soup):
        return (ColBasic.getTextFromElement(soup.select_one('h1')),
                ColBasic.cleanText(soup.select_one('#date_text').text.replace('입력 :', '').strip()[:17].strip()) if soup.select_one('#date_text') else None,
                ColBasic.getTextFromElement(soup.select_one('.par')))



__all__ = ["KoSymSpell"]

from re import I
import pkg_resources
import string
import unicodedata
from itertools import cycle
import math

from symspellpy import SymSpell, Verbosity
from symspellpy.symspellpy import SuggestItem, Composition
from symspellpy_ko.unicode import split_syllables, join_jamos


def get_resource(name):
    resource_package = __name__
    resource_path = "/".join(("res", name))
    filename = pkg_resources.resource_filename(resource_package, resource_path)
    return filename


class KoSymSpell(SymSpell):

    _decompose_korean = False

    def __init__(
        self, max_dictionary_edit_distance=2, prefix_length=7, count_threshold=1
    ):
        super().__init__(
            max_dictionary_edit_distance=max_dictionary_edit_distance,
            prefix_length=prefix_length,
            count_threshold=count_threshold,
        )

    def load_korean_dictionary(
        self, decompose_korean=True, load_bigrams=False, encoding="utf-8"
    ):
        self._decompose_korean = decompose_korean

        if decompose_korean:
            dict_path = "ko_50k_d.txt"
            bigram_path = "ko_bigrams_1M_d.txt"
        else:
            dict_path = "ko_50k.txt"
            bigram_path = "ko_bigrams_1M.txt"

        dict_path = get_resource(dict_path)
        bigram_path = get_resource(bigram_path)

        self.load_dictionary(dict_path, term_index=0, count_index=1, encoding=encoding)
        if load_bigrams:
            self.load_bigram_dictionary(
                bigram_path, term_index=0, count_index=2, encoding=encoding
            )

    def lookup(
        self,
        phrase,
        verbosity,
        max_edit_distance=None,
        include_unknown=False,
        ignore_token=None,
        transfer_casing=False,
    ):
        """Find suggested spellings for a given phrase word.

        한글 자소분해된 사전을 로드했을 경우 자소분해해주는 기능까지 추가함
        Parameters
        ----------
        phrase : str
            The word being spell checked.
        verbosity : :class:`Verbosity`
            The value controlling the quantity/closeness of the
            returned suggestions.
        max_edit_distance : int, optional
            The maximum edit distance between phrase and suggested
            words. Set to :attr:`_max_dictionary_edit_distance` by
            default
        include_unknown : bool, optional
            A flag to determine whether to include phrase word in
            suggestions, if no words within edit distance found.
        ignore_token : regex pattern, optional
            A regex pattern describing what words/phrases to ignore and
            leave unchanged
        transfer_casing : bool, optional
            A flag to determine whether the casing --- i.e., uppercase
            vs lowercase --- should be carried over from `phrase`.
        Returns
        -------
        suggestions : list
            suggestions is a list of :class:`SuggestItem` objects
            representing suggested correct spellings for the phrase
            word, sorted by edit distance, and secondarily by count
            frequency.
        Raises
        ------
        ValueError
            If `max_edit_distance` is greater than
            :attr:`_max_dictionary_edit_distance`
        """
        if self._decompose_korean:
            phrase = split_syllables(phrase)

        result = super().lookup(
            phrase,
            verbosity,
            max_edit_distance=max_edit_distance,
            include_unknown=include_unknown,
            ignore_token=ignore_token,
            transfer_casing=transfer_casing,
        )

        if self._decompose_korean:
            return [
                SuggestItem(join_jamos(sugg.term), sugg.distance, sugg.count)
                for sugg in result
            ]
        else:
            return result

    def lookup_compound(
        self,
        phrase,
        max_edit_distance,
        ignore_non_words=False,
        transfer_casing=False,
        split_phrase_by_space=False,
        ignore_term_with_digits=False,
    ):
        """`lookup_compound` supports compound aware automatic spelling
        correction of multi-word input strings with three cases:
        1. mistakenly inserted space into a correct word led to two
           incorrect terms
        2. mistakenly omitted space between two correct words led to
           one incorrect combined term
        3. multiple independent input terms with/without spelling
           errors
        Find suggested spellings for a multi-word input string
        (supports word splitting/merging).

        한글 자소분해된 사전을 로드했을 경우 자소분해해주는 기능까지 추가함

        Parameters
        ----------
        phrase : str
            The string being spell checked.
        max_edit_distance : int
            The maximum edit distance between input and suggested
            words.
        ignore_non_words : bool, optional
            A flag to determine whether numbers and acronyms are left
            alone during the spell checking process
        transfer_casing : bool, optional
            A flag to determine whether the casing --- i.e., uppercase
            vs lowercase --- should be carried over from `phrase`.
        split_by_space: bool, optional
            Splits the phrase into words simply based on space
        ignore_any_term_with_digits: bool, optional
            A flag to determine whether any term with digits
            is left alone during the spell checking process
        Returns
        -------
        suggestions_line : list
            suggestions_line is a list of :class:`SuggestItem` objects
            representing suggested correct spellings for `phrase`.
        """
        if self._decompose_korean:
            phrase = split_syllables(phrase)

        result = super().lookup_compound(
            phrase,
            max_edit_distance,
            ignore_non_words=ignore_non_words,
            transfer_casing=transfer_casing,
            split_phrase_by_space=split_phrase_by_space,
            ignore_term_with_digits=ignore_term_with_digits,
        )

        if self._decompose_korean:
            return [
                SuggestItem(join_jamos(sugg.term), sugg.distance, sugg.count)
                for sugg in result
            ]
        else:
            return result

    def word_segmentation(
        self,
        phrase,
        max_edit_distance=None,
        max_segmentation_word_length=None,
        ignore_token=None,
    ):
        """`word_segmentation` divides a string into words by inserting
        missing spaces at the appropriate positions misspelled words
        are corrected and do not affect segmentation existing spaces
        are allowed and considered for optimum segmentation
        `word_segmentation` uses a novel approach *without* recursion.
        https://medium.com/@wolfgarbe/fast-word-segmentation-for-noisy-text-2c2c41f9e8da
        While each string of length n can be segmented in 2^n−1
        possible compositions
        https://en.wikipedia.org/wiki/Composition_(combinatorics)
        `word_segmentation` has a linear runtime O(n) to find the optimum
        composition
        Find suggested spellings for a multi-word input string
        (supports word splitting/merging).

        자소분해한 한글을 합쳐버리기 때문에 원본 코드를 가져와서 수정함
        원본: https://github.com/mammothb/symspellpy/blob/e7a91a88f45dc4051b28b83e990fe072cabf0595/symspellpy/symspellpy.py#L946

        Parameters
        ----------
        phrase : str
            The string being spell checked.
        max_segmentation_word_length : int
            The maximum word length that should be considered.
        max_edit_distance : int, optional
            The maximum edit distance between input and corrected words
            (0=no correction/segmentation only).
        ignore_token : regex pattern, optional
            A regex pattern describing what words/phrases to ignore and
            leave unchanged
        Returns
        -------
        compositions[idx] :class:`Composition`
            The word segmented string, the word segmented and spelling
            corrected string, the edit distance sum between input
            string and corrected string, the sum of word occurence
            probabilities in log scale (a measure of how common and
            probable the corrected segmentation is).
        """

        # normalize ligatures: scientiﬁc -> scientific
        # 이 부분에서 한글 자소분해를 한게 합쳐져버려서 원본 코드를 가져와서 분해하게했음.
        phrase = unicodedata.normalize("NFKC", phrase).replace("\u002D", "")
        if self._decompose_korean:
            phrase = split_syllables(phrase)

        if max_edit_distance is None:
            max_edit_distance = self._max_dictionary_edit_distance
        if max_segmentation_word_length is None:
            max_segmentation_word_length = self._max_length
        array_size = min(max_segmentation_word_length, len(phrase))
        compositions = [Composition()] * array_size
        circular_index = cycle(range(array_size))
        idx = -1

        # outer loop (column): all possible part start positions
        for j in range(len(phrase)):
            # inner loop (row): all possible part lengths (from start
            # position): part can't be bigger than longest word in
            # dictionary (other than long unknown word)
            imax = min(len(phrase) - j, max_segmentation_word_length)
            for i in range(1, imax + 1):
                # get top spelling correction/ed for part
                part = phrase[j : j + i]
                separator_len = 0
                top_ed = 0
                top_log_prob = 0.0
                top_result = ""

                if part[0].isspace():
                    # remove space for levensthein calculation
                    part = part[1:]
                else:
                    # add ed+1: space did not exist, had to be inserted
                    separator_len = 1

                # remove space from part1, add number of removed spaces
                # to top_ed
                top_ed += len(part)
                # remove space.
                # add number of removed spaces to ed
                part = part.replace(" ", "")
                top_ed -= len(part)

                # v6.7: Lookup against the lowercase term
                results = self.lookup(
                    part.lower(),
                    Verbosity.TOP,
                    max_edit_distance,
                    ignore_token=ignore_token,
                )
                if results:
                    top_result = results[0].term
                    # v6.7: retain/preserve upper case
                    if len(part) > 0 and part[0].isupper():
                        top_result = top_result.capitalize()

                    top_ed += results[0].distance
                    # Naive Bayes Rule. We assume the word
                    # probabilities of two words to be independent.
                    # Therefore the resulting probability of the word
                    # combination is the product of the two word
                    # probabilities. Instead of computing the product
                    # of probabilities we are computing the sum of the
                    # logarithm of probabilities because the
                    # probabilities of words are about 10^-10, the
                    # product of many such small numbers could exceed
                    # (underflow) the floating number range and become
                    # zero. log(ab)=log(a)+log(b)
                    top_log_prob = math.log10(float(results[0].count) / float(self.N))
                else:
                    top_result = part
                    # default, if word not found. otherwise long input
                    # text would win as long unknown word (with
                    # ed=edmax+1), although there there should many
                    # spaces inserted
                    top_ed += len(part)
                    top_log_prob = math.log10(10.0 / self.N / math.pow(10.0, len(part)))

                dest = (i + idx) % array_size
                # set values in first loop
                if j == 0:
                    compositions[dest] = Composition(
                        part, top_result, top_ed, top_log_prob
                    )
                # pylint: disable=C0301,R0916
                elif (
                    i == max_segmentation_word_length
                    # replace values if better log_prob_sum, if same
                    # edit distance OR one space difference
                    or (
                        (
                            compositions[idx].distance_sum + top_ed
                            == compositions[dest].distance_sum
                            or compositions[idx].distance_sum + separator_len + top_ed
                            == compositions[dest].distance_sum
                        )
                        and compositions[dest].log_prob_sum
                        < compositions[idx].log_prob_sum + top_log_prob
                    )
                    # replace values if smaller edit distance
                    or compositions[idx].distance_sum + separator_len + top_ed
                    < compositions[dest].distance_sum
                ):
                    if (
                        len(top_result) == 1 and top_result[0] in string.punctuation
                    ) or (len(top_result) == 2 and top_result.startswith("'")):
                        compositions[dest] = Composition(
                            compositions[idx].segmented_string + part,
                            compositions[idx].corrected_string + top_result,
                            compositions[idx].distance_sum + top_ed,
                            compositions[idx].log_prob_sum + top_log_prob,
                        )
                    else:
                        compositions[dest] = Composition(
                            compositions[idx].segmented_string + " " + part,
                            (compositions[idx].corrected_string + " " + top_result),
                            (compositions[idx].distance_sum + separator_len + top_ed),
                            compositions[idx].log_prob_sum + top_log_prob,
                        )
            idx = next(circular_index)

        result = compositions[idx]

        if self._decompose_korean:
            return Composition(
                result.segmented_string,
                join_jamos(result.corrected_string),
                result.distance_sum,
                result.log_prob_sum,
            )
        else:
            return result



class BaseChunker:
    IN_TYPE = [str]
    OUT_TYPE = [list, str]

class KoreanChunker(BaseChunker):
    def __init__(self):

        import nltk
        grammar = """
        NP: {<N.*>*<Suffix>?}   # Noun phrase
        VP: {<V.*>*}            # Verb phrase
        AP: {<A.*>*}            # Adjective phrase
        """

        self.inst=nltk.RegexpParser(grammar)


    def __call__(self, *args, **kwargs):
        import konlpy
        words = konlpy.tag.Komoran().pos(*args)

        chunks = self.inst.parse(words)

        return chunks